###############################################################################
## Respawn and/or Teleport to predefined locations on '/respawn' command

###############################################################################
## Imports

from __future__ import with_statement
import es
import cmdlib
import os
import pickle

###############################################################################
## Info

info = es.AddonInfo()
info.author = 'me'
info.name = 'respawnto'
info.basename = 'respawnto'
info.version = '0.1'
info.url = 'http://addons.eventscripts.com/view/respawnto'
info.description = 'respawn players to pre-defined locations per map'

###############################################################################
## Globals

FILEREF_DB = os.path.join(es.getAddonPath(info.basename),
    'respawn_locations.db'
)
SPAWN_CT = 0
SPAWN_T = 1
SPAWN_ALL = 2
PLAYER_TERRORIST = 2
PLAYER_COUNTER_TERRORIST = 3
PLAYER_ALIVE = 0
PLAYER_DEAD = 1

###############################################################################
## Base classes

class Singleton(object):
	_instance = None
	def __new__(self, *args, **kwargs):
		if not self._instance:
			self._instance = super(Singleton, self).__new__(
				self, *args, **kwargs
			)
		return self._instance

###############################################################################
## Main classes

class Location(object):
    loc = None
    ang = None

class RespawnLocations(Singleton):
    _initialized = False
    _locations = None
    
    def __init__(self):
        if not self._initialized:
            self._initialized = True
            if os.path.isfile(FILEREF_DB):
                with open(FILEREF_DB, 'r') as f:
                    self._locations = pickle.load(f)
            else:
                self._locations = {}
                
    def exists(self, map, spawn_type):
        if not self._locations.has_key(map): return False
        else:
            if spawn_type == SPAWN_T:
                return self._locations[map].has_key('t')
            elif spawn_type == SPAWN_CT:
                return self._locations[map].has_key('ct')
            else:
                return (self._locations[map].has_key('ct') and
                    self._locations[map].has_key('t'))
                
    def get_location(self, map, spawn_type):
        if not self._locations.has_key(map): return None
        else:
            if spawn_type == SPAWN_T or spawn_type == SPAWN_ALL:
                if self._locations[map].has_key('t'):
                    return self._locations[map]['t']
                else: return None
            elif spawn_type == SPAWN_CT:
                if self._locations[map].has_key('ct'):
                    return self._locations[map]['ct']
                else: return None
            else: return None
            
    def set_location(self, map, spawn_type, loc):
        if self._locations.has_key(map):
            if spawn_type == SPAWN_T or spawn_type == SPAWN_ALL:
                self._locations[map]['t'] = loc
            if spawn_type == SPAWN_CT or spawn_type == SPAWN_ALL:
                self._locations[map]['ct'] = loc
        else:
            if spawn_type == SPAWN_T:
                self._locations[map] = { 't' : loc, }
            elif spawn_type == SPAWN_CT:
                self._locations[map] = { 'ct' : loc, }
            elif spawn_type == SPAWN_ALL:
                self._locations[map] = { 't' : loc, 'ct' : loc, }
            else: return
            
    def save_db(self):
        with open(FILEREF_DB, 'w') as f:
            pickle.dump(self._locations, f)
            
    def nice_boat(self):
        self.save_db()
    
class RespawnManager(Singleton):
    _initialized = False
    _current_map = None

    def __init__(self):
        if not self._initialized:
            self._initialized = True
            self.register_admin_commands()
            self.register_player_commands()
            RespawnLocations()
            
    def register_admin_commands(self):
        try:
            cmdlib.registerSayCommand('/set_respawn_point_t',
                saycmd_set_respawn_point_t, '', 'set_respawn_point', 'ADMIN'
            )
            cmdlib.registerSayCommand('/set_respawn_point_ct',
                saycmd_set_respawn_point_ct, '', 'set_respawn_point', 'ADMIN'
            )
            cmdlib.registerSayCommand('/set_respawn_point_all',
                saycmd_set_respawn_point_all, '', 'set_respawn_point', 'ADMIN'
            )
        except NameError: pass

    def register_player_commands(self):
        try:
            cmdlib.registerSayCommand('/respawn', saycmd_respawn_player, '')
        except NameError: pass
        
    def set_current_map(self, map):
        if not self._current_map == map:
         self._current_map = map
         
    def get_current_map(self):
        return self._current_map

    def spawn_player(self, userid):
        player_team = es.getplayerteam(userid)
        if (player_team == PLAYER_TERRORIST or
        player_team == PLAYER_COUNTER_TERRORIST):
            if not is_player_alive(userid):
                respawn_player(userid)
            if player_team == PLAYER_TERRORIST:
                loc = RespawnLocations().get_location(self.get_current_map(),
                    SPAWN_T
                )
                if loc:
                    teleport_player(userid, loc.loc, loc.ang)
            else:
                loc = RespawnLocations().get_location(self.get_current_map(),
                    SPAWN_CT
                )
                if loc:
                    teleport_player(userid, loc.loc, loc.ang)

    def set_spawn_point(self, userid, spawn_type):
        loc = Location()
        loc.loc = es.getplayerlocation(userid)
        loc.ang = get_player_ang(userid)
        if spawn_type == SPAWN_T or spawn_type == SPAWN_ALL:
            RespawnLocations().set_location(self.get_current_map(),
                spawn_type, loc
            )
        else:
            RespawnLocations().set_location(self.get_current_map(),
                spawn_type, loc
            )

    def nice_boat(self):
        RespawnLocations().nice_boat()
        cmdlib.unregisterSayCommand('/set_respawn_point_t')
        cmdlib.unregisterSayCommand('/set_respawn_point_ct')
        cmdlib.unregisterSayCommand('/set_respawn_point_all')
        cmdlib.unregisterSayCommand('/respawn')
        
###############################################################################
## Callbacks

def saycmd_set_respawn_point_t(userid, args):
    if is_player_alive(userid):
        RespawnManager().set_spawn_point(userid, SPAWN_T)

def saycmd_set_respawn_point_ct(userid, args):
    if is_player_alive(userid):
        RespawnManager().set_spawn_point(userid, SPAWN_CT)

def saycmd_set_respawn_point_all(userid, args):
    if is_player_alive(userid):
        RespawnManager().set_spawn_point(userid, SPAWN_ALL)

def saycmd_respawn_player(userid, args):
    RespawnManager().spawn_player(userid)

###############################################################################
## EventScripts events

def load():
    RespawnManager().set_current_map(str(
        es.ServerVar('eventscripts_currentmap')
    ))

def unload():
    RespawnManager().nice_boat()

def player_spawn(e):
    userid = int(e['userid'])
    player_team = es.getplayerteam(userid)
    if player_team == PLAYER_TERRORIST: player_team = SPAWN_T
    elif player_team == PLAYER_COUNTER_TERRORIST: player_team = SPAWN_CT
    else: return
    if not RespawnLocations().exists(RespawnManager().get_current_map(),
        player_team
    ):
        RespawnManager().set_spawn_point(userid, player_team)

def es_map_start(e):
    RespawnLocations().save_db()
    RespawnManager().set_current_map(e['mapname'])
    
###############################################################################
## Helper funcs

def is_player_alive(userid):
    dead_flag = int(es.getplayerprop(userid, 'CBasePlayer.pl.deadflag'))
    if dead_flag == PLAYER_ALIVE: return True
    else: return False
    
def get_player_ang(userid):
    return (float(es.getplayerprop(userid, 'CCSPlayer.m_angEyeAngles[0]')),
        float(es.getplayerprop(userid, 'CCSPlayer.m_angEyeAngles[1]')), 0.0,
    )
    
def respawn_player(userid):
    es.setplayerprop(userid, 'CCSPlayer.m_iPlayerState', 0)
    es.setplayerprop(userid, 'CCSPlayer.baseclass.m_lifeState', 512)
    es.spawnplayer(userid)
    
def teleport_player(userid, loc, ang):
    vel_x = es.getplayerprop(userid, 'CBasePlayer.localdata.m_vecVelocity[0]')
    vel_y = es.getplayerprop(userid, 'CBasePlayer.localdata.m_vecVelocity[1]')
    vel_z = es.getplayerprop(userid, 'CBasePlayer.localdata.m_vecVelocity[2]')
    negation = es.createvectorstring(float(vel_x) * -1.0,
        float(vel_y) * -1.0, float(vel_z) * -1.0
    )
    es.setplayerprop(userid, 'CBasePlayer.localdata.m_vecBaseVelocity',
        negation
    )
    es.server.cmd('es_xsetpos %s %s %s %s' %
        (userid, loc[0], loc[1], loc[2])
    )
    es.server.cmd('es_xsetang %s %s %s %s' %
        (userid, ang[0], ang[1], ang[2])
    )
